@@ -1,197 +1,18 @@
-import os, sys, psutil, time
+import sys, time
 import numpy as np
 import dask.array as da
 import SimpleITK as sitk
 from ClusterWrap.decorator import cluster
 import CircuitSeeker.utility as ut
+from CircuitSeeker._wrap_irm import configure_irm
 from CircuitSeeker.transform import apply_transform
 from CircuitSeeker.transform import compose_affine_and_displacement_vector_field
 from CircuitSeeker.transform import compose_displacement_vector_fields
 from CircuitSeeker.quality import jaccard_filter
+from CircuitSeeker.metrics import patch_mutual_information
 from scipy.spatial.transform import Rotation
 
 
-def configure_irm(
-    metric='MI',
-    bins=128,
-    sampling='regular',
-    sampling_percentage=1.0,
-    optimizer='GD',
-    iterations=200,
-    learning_rate=1.0,
-    estimate_learning_rate="once",
-    min_step=0.1,
-    max_step=1.0,
-    shrink_factors=[2,1],
-    smooth_sigmas=[2.,1.],
-    num_steps=[2, 2, 2],
-    step_sizes=[1., 1., 1.],
-    callback=None,
-):
-    """
-    Wrapper exposing some of the itk::simple::ImageRegistrationMethod API
-    Rarely called by the user. Typically used in custom registration functions.
-
-    Parameters
-    ----------
-    metric : string (default: 'MI')
-        The image matching term optimized during alignment
-        Options:
-            'MI': mutual information
-            'CC': correlation coefficient
-            'MS': mean squares
-
-    bins : int (default: 128)
-        Only used when `metric`='MI'. Number of histogram bins
-        for image intensity histograms. Ignored when `metric` is
-        'CC' or 'MS'
-
-    sampling : string (default: 'regular')
-        How image intensities are sampled during metric calculation
-        Options:
-            'regular': sample intensities with regular spacing
-            'random': sample intensities randomly
-
-    sampling_percentage : float in range [0., 1.] (default: 1.0)
-        Percentage of voxels used during metric sampling
-
-    optimizer : string (default 'GD')
-        Optimization algorithm used to find a transform
-        Options:
-            'GD': gradient descent
-            'RGD': regular gradient descent
-            'EX': exhaustive - regular sampling of transform parameters between
-                  given limits
-
-    iterations : int (default: 200)
-        Maximum number of iterations at each scale level to run optimization.
-        Optimization may still converge early.
-
-    learning_rate : float (default: 1.0)
-        Initial gradient descent step size
-
-    estimate_learning_rate : string (default: "once")
-        Frequency of estimating the learning rate. Only used if `optimizer`='GD'
-        Options:
-            'once': only estimate once at the beginning of optimization
-            'each_iteration': estimate step size at every iteration
-            'never': never estimate step size, `learning_rate` is fixed
-
-    min_step : float (default: 0.1)
-        Minimum allowable gradient descent step size. Only used if `optimizer`='RGD'
-
-    max_step : float (default: 1.0)
-        Maximum allowable gradient descent step size. Used by both 'GD' and 'RGD'
-
-    shrink_factors : iterable of type int (default: [2, 1])
-        Downsampling scale levels at which to optimize
-
-    smooth_sigmas : iterable of type float (default: [2., 1.])
-        Sigma of Gaussian used to smooth each scale level image
-        Must be same length as `shrink_factors`
-        Should be specified in physical units, e.g. mm or um
-
-    num_steps : iterable of type int (default: [2, 2, 2])
-        Only used if `optimizer`='EX'
-        Number of steps to search in each direction from the initial
-        position of the transform parameters
-
-    step_sizes : iterable of type float (default: [1., 1., 1.])
-        Only used if `optimizer`='EX'
-        Size of step to take during brute force optimization
-        Order of parameters and relevant scales should be based on
-        the type of transform being optimized
-
-    callable : callable object, e.g. function (default: None)
-        A function run at every iteration of optimization
-        Should take only the ImageRegistrationMethod object as input: `irm`
-        If None then the Level, Iteration, and Metric values are
-        printed at each iteration
-
-    Returns
-    -------
-    irm : itk::simple::ImageRegistrationMethod object
-        The configured ImageRegistrationMethod object. Simply needs
-        images and a transform type to be ready for optimization.
-    """
-
-    # identify number of cores available, assume hyperthreading
-    if "LSB_DJOB_NUMPROC" in os.environ:
-        ncores = int(os.environ["LSB_DJOB_NUMPROC"])
-    else:
-        ncores = psutil.cpu_count(logical=False)
-
-    # initialize IRM object, be completely sure nthreads is set
-    sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(2*ncores)
-    irm = sitk.ImageRegistrationMethod()
-    irm.SetNumberOfThreads(2*ncores)
-
-    # set interpolator
-    irm.SetInterpolator(sitk.sitkLinear)
-
-    # set metric
-    if metric == 'MI':
-        irm.SetMetricAsMattesMutualInformation(
-            numberOfHistogramBins=bins,
-        )
-    elif metric == 'CC':
-        irm.SetMetricAsCorrelation()
-    elif metric == 'MS':
-        irm.SetMetricAsMeanSquares()
-
-    # set metric sampling type and percentage
-    if sampling == 'regular':
-        irm.SetMetricSamplingStrategy(irm.REGULAR)
-    elif sampling == 'random':
-        irm.SetMetricSamplingStrategy(irm.RANDOM)
-    irm.SetMetricSamplingPercentage(sampling_percentage)
-
-    # set estimate learning rate
-    if estimate_learning_rate == "never":
-        estimate_learning_rate = irm.Never
-    elif estimate_learning_rate == "once":
-        estimate_learning_rate = irm.Once
-    elif estimate_learning_rate == "each_iteration":
-        estimate_learning_rate = irm.EachIteration
-
-    # set optimizer
-    if optimizer == 'GD':
-        irm.SetOptimizerAsGradientDescent(
-            numberOfIterations=iterations,
-            learningRate=learning_rate,
-            maximumStepSizeInPhysicalUnits=max_step,
-            estimateLearningRate=estimate_learning_rate,
-        )
-        irm.SetOptimizerScalesFromPhysicalShift()
-    elif optimizer == 'RGD':
-        irm.SetOptimizerAsRegularStepGradientDescent(
-            minStep=min_step, learningRate=learning_rate,
-            numberOfIterations=iterations,
-            maximumStepSizeInPhysicalUnits=max_step,
-        )
-        irm.SetOptimizerScalesFromPhysicalShift()
-    elif optimizer == 'EX':
-        irm.SetOptimizerAsExhaustive(num_steps[::-1])
-        irm.SetOptimizerScales(step_sizes[::-1])
-
-    # set pyramid
-    irm.SetShrinkFactorsPerLevel(shrinkFactors=shrink_factors)
-    irm.SetSmoothingSigmasPerLevel(smoothingSigmas=smooth_sigmas)
-    irm.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()
-
-    # set callback function
-    if callback is None:
-        def callback(irm):
-            level = irm.GetCurrentLevel()
-            iteration = irm.GetOptimizerIteration()
-            metric = irm.GetMetricValue()
-            print("LEVEL: ", level, " ITERATION: ", iteration, " METRIC: ", metric)
-    irm.AddCommand(sitk.sitkIterationEvent, lambda: callback(irm))
-
-    # return configured irm
-    return irm
-
-
 def random_affine_search(
     fix,
     mov,
@@ -209,6 +30,8 @@ def random_affine_search(
     fix_origin=None,
     mov_origin=None,
     jaccard_filter_threshold=None,
+    use_patch_mutual_information=False,
+    print_running_improvements=False,
     **kwargs,
 ):
     """
@@ -347,9 +170,6 @@ def random_affine_search(
     if max_scale != 1: params[1:, 6:9] = np.e**F(np.log(max_scale))
     if max_shear != 0: params[1:, 9:] = F(max_shear)
 
-    # set up registration object
-    irm = configure_irm(**kwargs)
-
     # skip sample to alignment spacing
     if alignment_spacing is not None:
         fix, fix_spacing_ss = ut.skip_sample(fix, fix_spacing, alignment_spacing)
@@ -361,36 +181,93 @@ def random_affine_search(
         fix_spacing = fix_spacing_ss
         mov_spacing = mov_spacing_ss
 
-    # convert to float32 sitk images
-    fix_sitk = ut.numpy_to_sitk(fix, fix_spacing, origin=fix_origin)
-    mov_sitk = ut.numpy_to_sitk(mov, mov_spacing, origin=mov_origin)
-    fix_sitk = sitk.Cast(fix_sitk, sitk.sitkFloat32)
-    mov_sitk = sitk.Cast(mov_sitk, sitk.sitkFloat32)
+    # define metric evaluation
+    if use_patch_mutual_information:
 
-    # set masks
-    if fix_mask is not None:
-        fix_mask_sitk = ut.numpy_to_sitk(fix_mask, fix_spacing, origin=fix_origin)
-        irm.SetMetricFixedMask(fix_mask_sitk)
-    if mov_mask is not None:
-        mov_mask_sitk = ut.numpy_to_sitk(mov_mask, mov_spacing, origin=mov_origin)
-        irm.SetMetricMovingMask(mov_mask_sitk)
+        # wrap patch_mi metric
+        def score_affine(affine):
+
+            # apply transform
+            aligned = apply_transform(
+                fix, mov, fix_spacing, mov_spacing,
+                transform_list=[affine,],
+                fix_origin=fix_origin,
+                mov_origin=mov_origin,
+            )
+
+            # mov mask
+            mov_mask_aligned = None
+            if mov_mask is not None:
+                mov_mask_aligned = apply_transform(
+                    fix, mov_mask, fix_spacing, mov_spacing,
+                    transform_list=[affine,],
+                    fix_origin=fix_origin,
+                    mov_origin=mov_origin,
+                    interpolate_with_nn=True,
+                )
+
+            return patch_mutual_information(
+                fix, aligned, fix_spacing,
+                fix_mask=fix_mask,
+                mov_mask=mov_mask_aligned,
+                return_metric_image=False,
+                **kwargs,
+            )
+
+    # otherwise score entire image domain
+    else:
+
+        # convert to float32 sitk images
+        fix_sitk = ut.numpy_to_sitk(fix, fix_spacing, origin=fix_origin)
+        mov_sitk = ut.numpy_to_sitk(mov, mov_spacing, origin=mov_origin)
+        fix_sitk = sitk.Cast(fix_sitk, sitk.sitkFloat32)
+        mov_sitk = sitk.Cast(mov_sitk, sitk.sitkFloat32)
+
+        # create irm to use metric object
+        irm = configure_irm(**kwargs)
+
+        # set masks
+        if fix_mask is not None:
+            fix_mask_sitk = ut.numpy_to_sitk(fix_mask, fix_spacing, origin=fix_origin)
+            irm.SetMetricFixedMask(fix_mask_sitk)
+        if mov_mask is not None:
+            mov_mask_sitk = ut.numpy_to_sitk(mov_mask, mov_spacing, origin=mov_origin)
+            irm.SetMetricMovingMask(mov_mask_sitk)
+
+        # wrap full image mi
+        def score_affine(affine):
+
+            # reformat affine as tranfsorm and give to irm
+            affine = ut.matrix_to_affine_transform(affine)
+            irm.SetMovingInitialTransform(affine)
+
+            # get the metric value
+            try:
+                return irm.MetricEvaluate(fix_sitk, mov_sitk)
+            except Exception as e:
+                return np.finfo(scores.dtype).max
 
     # score all random affines
+    fail_count = 0
+    current_best_score = 0
     scores = np.empty(random_iterations + 1)
-    fail_count = 0  # keep track of failures
     for iii, ppp in enumerate(params):
         aff = params_to_affine_matrix(ppp)
-        aff = ut.matrix_to_affine_transform(aff)
-        irm.SetMovingInitialTransform(aff)
-        try:
-            scores[iii] = irm.MetricEvaluate(fix_sitk, mov_sitk)
-        except Exception as e:
-            scores[iii] = np.finfo(scores.dtype).max
+        scores[iii] = score_affine(aff)
+        if scores[iii] == np.finfo(scores.dtype).max:
             fail_count += 1
-            if fail_count >= 10 or fail_count >= random_iterations + 1:
-                print("Random search failed due to ITK exception:\n", e)
-                print("Returning default")
-                return np.eye(4)
+
+        # print running improvements
+        if print_running_improvements:
+            if scores[iii] < current_best_score:
+                current_best_score = scores[iii]
+                print(iii, ': ', current_best_score, '\n', aff, '\n', flush=True)
+
+        # check for excessive failure
+        if fail_count >= 10 or fail_count >= random_iterations + 1:
+            print("Random search failed due to ITK exceptions")
+            print("Returning default")
+            return np.eye(4)
 
     # sort
     params = params[np.argsort(scores)]
@@ -404,6 +281,8 @@ def random_affine_search(
         scores = np.empty(affine_align_best)
         fail_count = 0  # keep track of failures
         for iii in range(affine_align_best):
+
+            # gradient descent affine alignment
             aff = params_to_affine_matrix(params[iii])
             aff = affine_align(
                fix, mov, fix_spacing, mov_spacing,
@@ -415,17 +294,13 @@ def random_affine_search(
                alignment_spacing=None,  # already done in this function
                **kwargs,
             )
-            aff = ut.matrix_to_affine_transform(aff)
-            irm.SetMovingInitialTransform(aff)
-            try:
-                scores[iii] = irm.MetricEvaluate(fix_sitk, mov_sitk)
-            except Exception as e:
-                scores[iii] = np.finfo(scores.dtype).max
-                fail_count += 1
-                if fail_count >= affine_align_best:
-                    print("Random search failed due to ITK exception:\n", e)
-                    print("Returning default")
-                    return np.eye(4)
+
+            # score the result
+            scores[iii] = score_affine(aff)
+            if fail_count >= affine_align_best:
+                print("Random search failed due to ITK exceptions")
+                print("Returning default")
+                return np.eye(4)
 
         # return the best one
         return params_to_affine_matrix(params[np.argmin(scores)])
